C*****************************************************
C                                                    *
C     Copyright (c) 2001 by Northwestern University. *
C     All Rights Reserved                            *
C                                                    *
C*****************************************************
c
c     ----------------------------------------------------------------
c                   KNITRO DRIVER FOR CUTEr INTERFACE
c     ----------------------------------------------------------------
c
c     KNITRO is a code for solving large-scale nonlinear programming
c     problems of the form
c     
c       min         f(x)
c       subject to       
c                   cl(i)  = c_i(x)  =  cu(i),   i = 1,...,num_equal
c                   cl(j) <= c_j(x) <=  cu(j),   j = num_equal+1,...,m
c                   bl(k) <=   x(k) <=  bu(k),   k = 1,...,n
c     
c     The code implements an interior point algorithm with trust 
c     region techniques. It uses first and second derivatives 
c     (if available) of the function and constraints.
c     
c     This driver runs the CUTEr test problems whose suffix names are
c     *.SIF. 
c     
c     References:
c
c       [1] R.H. Byrd, M.E. Hribar, and J. Nocedal, "An interior point 
c       algorithm for large scale nonlinear programming", SIAM Journal 
c       on Optimization, 9, 4, pp. 877-900 (1999).
c         Note: this paper gives a description of the algorithm
c               implemented in KNITRO. Some changes have occurred since
c               then.
c
c       [2] R.H. Byrd, J.C. Gilbert, and J. Nocedal, "A trust region
c       method based on interior point techniques for nonlinear 
c       programming", Mathematical Programming A, 89:149-185
c       (2000).
c         Note: this paper provides a global convergence analysis.
c
c     **************
c
      program          driver
c
c     This driver uses reverse communication, and interfaces with 
c     CUTEr by calling the following subroutines: 
c
c       csetup ---- get the problem dimension n, the number m of 
c                   constraints, the bounds cl, cu, bl, bu. 
c 
c       cnames ---- get the problem name, problem_name.
c
c       cfn    ---- compute the objective function value f and the 
c                   constraints c.
c
c       csgr   ---- compute the gradient g of the objective function, 
c                   and the constraint gradients. 
c
c       csh    ---- compute the Hessian of the Lagrangian function.
c 
      implicit none

c**********************************************************************
c
c     We begin by defining some parameters.
c
c     These parameters can be made smaller or larger to adjust for
c     application problem and host machine memory sizes. Please see
c     the recommended sizes below.
c
      integer  max_n, max_m, max_nnzj, max_nnz_W, lwork, liwork

c**********************************************************************
c
c     Recommended maximum array sizes  
c
c        n: number of variables
c        m: number of constraints
c
c     *  very small scale problems   n/m <=     200
c     *  small scale problems        n/m <=   1 000
c     *  medium scale problems       n/m <=  10 000
c     *  large scale problems        n/m <=  50 000
c     *  very large scale problems   n/m <= 250 000
c
c**********************************************************************
c
c     Very small scale problems
c
c     parameter (max_n       = 200)
c     parameter (max_m       = 200)
c     parameter (max_nnzj    = 40000)
c     parameter (max_nnz_W   = 40000)
c     parameter (lwork       = 200000)
c     parameter (liwork      = 400000) 

c     Small scale problems
c
c     parameter (max_n       = 1000)
c     parameter (max_m       = 1000)
c     parameter (max_nnzj    = 100000)
c     parameter (max_nnz_W   = 100000)
c     parameter (lwork       = 500000)
c     parameter (liwork      = 1000000) 

c     Medium scale problems
c
      parameter (max_n       = 10000)
      parameter (max_m       = 10000)
      parameter (max_nnzj    = 200000)
      parameter (max_nnz_W   = 200000)
      parameter (lwork       = 1000000)
      parameter (liwork      = 2000000) 

c     Large scale problems
c
c     parameter (max_n       = 50000)
c     parameter (max_m       = 50000)
c     parameter (max_nnzj    = 1000000)
c     parameter (max_nnz_W   = 1000000)
c     parameter (lwork       = 5000000)
c     parameter (liwork      = 10000000) 

c     Very large scale problems
c
c     parameter (max_n       = 250000)
c     parameter (max_m       = 250000)
c     parameter (max_nnzj    = 5000000)
c     parameter (max_nnz_W   = 5000000)
c     parameter (lwork       = 25000000)
c     parameter (liwork      = 50000000) 
c
c**********************************************************************
c
c     Define some communication flag parameters.
c
      integer           initial, eval_fc, eval_gA,
     $                  eval_W, eval_x0, finished
      parameter        (initial      =  0,
     $                  eval_fc      =  1,
     $                  eval_gA      =  2,
     $                  eval_W       =  3,
     $                  eval_x0      =  4,
     $                  finished     =  5)
c
c     Declare some variables, whose detailed description is at 
c     the end of this file.
c
      logical           equatn(max_m), linear(max_m)
      integer           n, m, nnz_W, nnzj, W_col(max_nnz_W),
     $                  W_row(max_nnz_W), indvar(max_nnzj),
     $                  indfun(max_nnzj), status, info,
     $                  iwork(liwork), minwork
      real              CALLS( 7 ), CPU( 2 )
      character*20      problem_name, cdummy(max_n)
      double precision  f, x(max_n), c(max_m), cl(max_m), cu(max_m),
     $                  w(max_nnz_W), lambda(max_m+max_n),
     $                  ddummy(max_m), bl(max_n), bu(max_n),
     $                  cjac(max_nnzj), work(lwork), vector(max_n)
c
c     The following arrays are specified in the specs 
c     file 'knitro.spc'.
c
      logical           lspecs(10)
      integer           ispecs(10)
      double precision  dspecs(10)

c**********************************************************************
c
c     Read in the user options from the file 'knitro.spc'.
c
      call readspecs(ispecs, dspecs, lspecs)
c
c     Set up the problem. Get n, m, bl, bu, cl, cu.
c
      open(15, file='OUTSDIF.d', form='formatted')
      call csetup(15, 6, n, m, x, bl, bu, max_n, equatn, linear,
     $            ddummy, cl, cu, max_m, .TRUE., .FALSE., .FALSE.)
c
c     Get the problem name.
c
      call cnames(n, m, problem_name, cdummy, cdummy)
c
c     We need to call the CUTEr routines 'cdimsj' and 'cdimsh'
c     before calling 'knitros' for the first time to get the
c     values nnzj and nnz_W which are required on input.
c     (If we are not using the exact Hessian version of KNITRO 
c     then we do not call 'cdimsh' and nnz_W is defined below.)
c  
      CALL CDIMSJ( nnzj )
      if (ispecs(4) .eq. 1) then
c  
c     Specify nnz_W for exact Hessian version.
c
         CALL CDIMSH( nnz_W )
      elseif (ispecs(4) .eq. 2 .or. ispecs(4) .eq. 3) then
c  
c     Specify nnz_W for quasi-Newton version.
c
         nnz_W = (n**2 - n)/2 + n
      elseif (ispecs(4) .eq. 4 .or. ispecs(4) .eq. 5) then
c
c     Hessian-vector product version does not store Hessian.
c
         nnz_W = 0         
      endif

      status = initial  
c
c             << beginning of loop >>
c
 111  continue
c
c     Compute function values.
c
      if ((status .eq. eval_fc) .or. (status .eq. eval_x0)) then
         call cfn(n, m, x, f, max_m, c)
      endif
c
c     Compute gradients.
c
      if ((status .eq. eval_gA) .or. (status .eq. eval_x0)) then
         call csgr(n, m, .FALSE., 1, ddummy, x, nnzj, max_nnzj,
     $        cjac, indvar, indfun)
      endif
c
c     Compute the Hessian W of the Lagrangian function.
c
      if ((status .eq. eval_W) .and. (ispecs(4) .eq. 1)) then
         call csh(n, m, x, m, lambda, nnz_W, max_nnz_W, 
     $        W, W_row, W_col)
      endif
c
c     Call the Knitro solver.
c
      call knitros(n, m, f, x, bl, bu, c, cl, cu, equatn, linear,
     $            nnzj, cjac, indvar, indfun, lambda, nnz_W, W,
     $            W_row, W_col, vector, ispecs, dspecs, lspecs,
     $            info, liwork, iwork, lwork, work, minwork, status)
c
c     If the stopping test has not been satisfied, re-enter the loop.
c
      if (status .ne. finished) goto 111
c
c                << end of the loop>>
c
      CALL CREPRT( CALLS, CPU )
      WRITE ( ispecs(1), 2000 ) problem_name, n, m, CALLS(1),
     *     CALLS(2), CALLS(5), CALLS(6), info, f, CPU(1), CPU(2)
c
 2000 FORMAT( /, 24('*'), ' CUTEr statistics ', 24('*') //
     *     ,' Code used               :  KNITRO',    /
     *     ,' Problem                 :  ', A10,    /
     *     ,' # variables             =      ', I10 /
     *     ,' # constraints           =      ', I10 /
     *     ,' # objective functions   =      ', E15.7 /
     *     ,' # objective gradients   =      ', E15.7 / 
     *     ,' # constraints functions =      ', E15.7 /
     *     ,' # constraints gradients =      ', E15.7 /
     *     ' Exit code               =      ', I10 /
     *     ,' Final f                 = ', E15.7 /
     *     ,' Set up time             =      ', 0P, F10.2, ' seconds' /
     *     ' Solve time              =      ', 0P, F10.2, ' seconds' //
     *     66('*') / )

      stop
      end
c
c==================== The end of CUTEr driver =======================
c
c     --------------------------------------------------------------
c             DESCRIPTION OF THE VARIABLES IN KNITRO
c     --------------------------------------------------------------
c
c     n is an INTEGER variable denoting the number of variables.
c 
c     m is an INTEGER variable denoting the number of general
c        constraints.
c
c     f is a DOUBLE PRECISION variable which holds the value of the
c        objective function.
c
c     x is a DOUBLE PRECISION array of length n.  It is the solution
c        vector.
c
c     bl is a DOUBLE PRECISION array of length n. bl(i) is the lower 
c        bound of the i-th variable x(i). If there is no such bound,
c        set it to be the large negative number -1.0d+20.
c
c     bu is a DOUBLE PRECISION array of length n. bu(i) is the upper 
c        bound of the i-th variable x(i). If there is no such bound,
c        set it to be the large positive number 1.0d+20.
c
c     c is a DOUBLE PRECISION array of length m.  It contains the
c        general equality and inequality constraint values (it
c        excludes fixed variables and bound constraints).
c
c     cl is a DOUBLE PRECISION array of length m. cl(i) is 
c	 the lower bound of the i-th constraint c(i). If there 
c        is no such bound, set it to be the large negative 
c        number -1.0d+20.
c
c     cu is a DOUBLE PRECISION  array of length m. cu(i) is 
c	 the upper bound of the i-th constraint c(i). If there
c        is no such bound, set it to be the large positive 
c        number 1.0d+20.
c
c     equatn is a LOGICAL variable of length m, with equatn(i) 
c        indicating whether the i-th constraint is an equality 
c        constraint or not. 
c
c     linear is a LOGICAL variable of length m, with linear(i) 
c        indicating whether the i-th constraint is a linear 
c        constraint or not. 
c
c     nnzj is an INTEGER. It is the number of nonzeros in the Jacobian 
c        matrix cjac which contains the gradient of the objective
c        function f and the constraint gradients in sparse form.
c
c     cjac is a DOUBLE PRECISION array of length nnzj. The first part  
c        contains the nonzero elements of the gradient of the objective
c        function; the second part contains the nonzero elements of the
c        Jacobian of the constraints. 
c
c     indvar is an INTEGER array of length nnzj. It is the index of the
c        variables.
c
c     indfun is an INTEGER array of length nnzj. It is the indicator 
c        for the functions. If indfun(i)=0, it refers to the objective 
c        function. If indfun(i)=j, it refers to the j-th constraint.
c
c        indfun(i) and indvar(i) determine the row number
c        and the column number respectively of nonzero Jacobian 
c        element cjac(i).
c 
c     lambda is a DOUBLE PRECISION array of length m+n containing the 
c        Lagrange multiplier estimates. The first m components hold 
c        the multipliers for the general constraints and the last
c        n components hold the multipliers for the bounds.
c        The Lagrangian function (excluding bounds) is
c
c        L(x, lambda) = f(x) + lambda*c(x).
c
c     nnz_W is an INTEGER variable denoting the number of nonzero
c        elements in the upper triangle of the Hessian of the
c        Lagrangian function. 
c
c     W is a DOUBLE PRECISION array of dimension nnz_W containing the 
c        Hessian of the Lagrangian in sparse form: 
c
c        W = nabla^2_{xx} f(x) + lambda*nabla^2_{xx} c(x)
c
c        Only the upper triangle is stored.
c
c     W_row is an INTEGER array of length nnz_W. W_row(i) stores
c        the row number of the nonzero element W(i).
c
c     W_col is an INTEGER array of length nnz_W. W_col(i) stores the 
c        column number of the nonzero element W(i).
c
c     vector is a DOUBLE PRECISION array of dimension n which is only
c        used if the user is providing a subroutine for computing
c        exact Hessian-vector products (option ispecs(4)=5).  In 
c        this case, on return from subroutine knitros vector holds
c        the vector which will be multiplied by the Hessian and on
c        re-entry vector must hold the desired Hessian-vector product.
c
c     ispecs is an INTEGER array of length 10 which holds the integer
c        user options. 
c        See the User's Manual for more detail.
c
c     dspecs is a DOUBLE PRECISION array of length 10 which holds the
c        double precision user options.
c        See the User's Manual for more detail.
c
c     lspecs is a LOGICAL array of length 10 which holds the
c        logical user options.
c        See the User's Manual for more detail.
c
c     info is an INTEGER variable which indicates the exit status.
c        info = 0: termination without error
c             < 0: termination with an error
c
c     liwork is an INTEGER variable which gives the length of the
c        integer work array, iwork. 
c
c     iwork is an INTEGER array of length liwork which is used
c        as an integer work vector.
c
c     lwork is an INTEGER variable which gives the length of the  
c        real work array, work. 
c
c     work is an DOUBLE PRECISION array of length lwork which is used
c        as a real work vector.
c
c     minwork is an INTEGER variable which gives some information
c        about the minimum work array sizes needed when these
c        arrays are too small.
c
c        info = -48: minwork gives minimum value for liwork
c        info = -49: minwork gives minimum value for lwork
c
c     status is an INTEGER which keeps track of the reverse 
c        communication status.  On initial entry status must
c        be set to 0.
c
c     --------------------------------------------------------------
c           END OF THE DESCRIPTION OF THE VARIABLES IN KNITRO
c     --------------------------------------------------------------
